@@ -6,14 +6,20 @@ import itertools
 import subprocess
 from pathlib import Path
 from contextlib import contextmanager
+from datetime import datetime
 from typing import Iterator, Union, List, BinaryIO
 
+try:
+    from collections.abc import Iterable
+except ImportError:
+    from collections import Iterable
+
 from python_pachyderm.mixin.pfs import PFSFile, PFSTarFile
 from python_pachyderm.experimental.pfs import commit_from, Commit, uuid_re
 from python_pachyderm.service import Service, pfs_proto as pfs_proto_pb
 from python_pachyderm.experimental.service import pfs_proto
 from python_pachyderm.experimental.util import check_pachctl
-from google.protobuf import wrappers_pb2
+from google.protobuf import wrappers_pb2, timestamp_pb2
 import betterproto.lib.google.protobuf as bp_proto
 
 # bp_to_pb: bp_proto.Empty -> empty_pb2.Empty
@@ -27,7 +33,11 @@ class PFSMixin:
     """A mixin with pfs-related functionality."""
 
     def create_repo(
-        self, repo_name: str, description: str = None, update: bool = False
+        self,
+        repo_name: str,
+        description: str = None,
+        update: bool = False,
+        project_name: str = None,
     ) -> None:
         """Creates a new repo object in PFS with the given name. Repos are the
         top level data object in PFS and should be used to store data of a
@@ -43,22 +53,32 @@ class PFSMixin:
             Description of the repo.
         update : bool, optional
             Whether to update if the repo already exists.
+        project_name : str
+            The name of the project.
         """
         self._req(
             Service.PFS,
             "CreateRepo",
-            repo=pfs_proto.Repo(name=repo_name, type="user"),
+            repo=pfs_proto.Repo(
+                name=repo_name,
+                type="user",
+                project=pfs_proto.Project(name=project_name),
+            ),
             description=description,
             update=update,
         )
 
-    def inspect_repo(self, repo_name: str) -> pfs_proto.RepoInfo:
+    def inspect_repo(
+        self, repo_name: str, project_name: str = None
+    ) -> pfs_proto.RepoInfo:
         """Inspects a repo.
 
         Parameters
         ----------
         repo_name : str
             Name of the repo.
+        project_name : str
+            The name of the project.
 
         Returns
         -------
@@ -66,10 +86,18 @@ class PFSMixin:
             A protobuf object with info on the repo.
         """
         return self._req(
-            Service.PFS, "InspectRepo", repo=pfs_proto.Repo(name=repo_name, type="user")
+            Service.PFS,
+            "InspectRepo",
+            repo=pfs_proto.Repo(
+                name=repo_name,
+                type="user",
+                project=pfs_proto.Project(name=project_name),
+            ),
         )
 
-    def list_repo(self, type: str = "user") -> Iterator[pfs_proto.RepoInfo]:
+    def list_repo(
+        self, type: str = "user", projects_filter: List[pfs_proto.Project] = None
+    ) -> Iterator[pfs_proto.RepoInfo]:
         """Lists all repos in PFS.
 
         Parameters
@@ -77,15 +105,22 @@ class PFSMixin:
         type : str, optional
             The type of repos that should be returned ("user", "meta", "spec").
             If unset, returns all types of repos.
+        projects_filter : [Project], optional
+            Filters out repos that do not belong in the list,
+            while no projects means to list all repos
 
         Returns
         -------
         Iterator[pfs_proto.RepoInfo]
             An iterator of protobuf objects that contain info on a repo.
         """
-        return self._req(Service.PFS, "ListRepo", type=type)
+        if isinstance(projects_filter, Iterable):
+            projects_filter = [pfs_proto.Project(name=p.name) for p in projects_filter]
+        return self._req(Service.PFS, "ListRepo", type=type, projects=projects_filter)
 
-    def delete_repo(self, repo_name: str, force: bool = False) -> None:
+    def delete_repo(
+        self, repo_name: str, force: bool = False, project_name: str = None
+    ) -> None:
         """Deletes a repo and reclaims the storage space it was using.
 
         Parameters
@@ -95,11 +130,17 @@ class PFSMixin:
         force : bool, optional
             If set to true, the repo will be removed regardless of errors.
             Use with care.
+        project_name : str
+            The name of the project.
         """
         self._req(
             Service.PFS,
             "DeleteRepo",
-            repo=pfs_proto.Repo(name=repo_name, type="user"),
+            repo=pfs_proto.Repo(
+                name=repo_name,
+                type="user",
+                project=pfs_proto.Project(name=project_name),
+            ),
             force=force,
         )
 
@@ -107,12 +148,84 @@ class PFSMixin:
         """Deletes all repos."""
         self._req(Service.PFS, "DeleteAll", req=bp_proto.Empty())
 
+    def create_project(
+        self, project_name: str, description: str = None, update: bool = False
+    ) -> None:
+        """Creates a new project with the given name.
+
+        Parameters
+        ----------
+        project_name : str
+            Name of the project.
+        description : str, optional
+            Description of the project.
+        update : bool, optional
+            Whether to update if the project already exists.
+        """
+        self._req(
+            Service.PFS,
+            "CreateProject",
+            project=pfs_proto.Project(name=project_name),
+            description=description,
+            update=update,
+        )
+
+    def inspect_project(self, project_name: str) -> pfs_proto.ProjectInfo:
+        """Inspects a project.
+
+        Parameters
+        ----------
+        project_name : str
+            Name of the project.
+
+        Returns
+        -------
+        pfs_proto.ProjectInfo
+            A protobuf object with info on the project.
+        """
+        return self._req(
+            Service.PFS, "InspectProject", project=pfs_proto.Project(name=project_name)
+        )
+
+    def list_project(self) -> Iterator[pfs_proto.ProjectInfo]:
+        """Lists all projects in PFS.
+
+        Returns
+        -------
+        Iterator[pfs_proto.ProjectInfo]
+            An iterator of protobuf objects that contain info on a project.
+        """
+        return self._req(
+            Service.PFS,
+            "ListProject",
+            req=bp_proto.Empty(),
+        )
+
+    def delete_project(self, project_name: str, force: bool = False) -> None:
+        """Deletes a project and reclaims the storage space it was using.
+
+        Parameters
+        ----------
+        project_name : str
+            The name of the project.
+        force : bool, optional
+            If set to true, the repo will be removed regardless of errors.
+            Use with care.
+        """
+        self._req(
+            Service.PFS,
+            "DeleteProject",
+            force=force,
+            project=pfs_proto.Project(name=project_name),
+        )
+
     def start_commit(
         self,
         repo_name: str,
         branch: str,
         parent: Union[str, tuple, dict, Commit, pfs_proto.Commit] = None,
         description: str = None,
+        project_name: str = None,
     ) -> pfs_proto.Commit:
         """Begins the process of committing data to a repo. Once started you
         can write to the commit with ModifyFile. When all the data has been
@@ -131,6 +244,8 @@ class PFSMixin:
             identical to the parent.
         description : str, optional
             A description of the commit.
+        project_name : str
+            The name of the project.
 
         Returns
         -------
@@ -142,20 +257,19 @@ class PFSMixin:
         --------
         >>> c = client.start_commit("foo", "master", ("foo", "staging"))
         """
+        repo = pfs_proto.Repo(
+            name=repo_name, type="user", project=pfs_proto.Project(name=project_name)
+        )
         if parent and isinstance(parent, str):
             parent = pfs_proto.Commit(
                 id=parent,
-                branch=pfs_proto.Branch(
-                    repo=pfs_proto.Repo(name=repo_name, type="user"), name=None
-                ),
+                branch=pfs_proto.Branch(repo=repo, name=None),
             )
         return self._req(
             Service.PFS,
             "StartCommit",
             parent=commit_from(parent),
-            branch=pfs_proto.Branch(
-                repo=pfs_proto.Repo(name=repo_name, type="user"), name=branch
-            ),
+            branch=pfs_proto.Branch(repo=repo, name=branch),
             description=description,
         )
 
@@ -213,6 +327,7 @@ class PFSMixin:
         branch: str,
         parent: Union[str, tuple, dict, Commit, pfs_proto.Commit] = None,
         description: str = None,
+        project_name: str = None,
     ) -> Iterator[pfs_proto.Commit]:
         """A context manager for running operations within a commit.
 
@@ -220,7 +335,7 @@ class PFSMixin:
         ----------
         repo_name : str
             The name of the repo.
-        branch_name : str
+        branch : str
             A string specifying the branch.
         parent : Union[str, tuple, dict, Commit, pfs_proto.Commit], optional
             A commit specifying the parent of the newly created commit. Upon
@@ -228,6 +343,8 @@ class PFSMixin:
             identical to the parent.
         description : str, optional
             A description of the commit.
+        project_name : str
+            The name of the project.
 
         Yields
         -------
@@ -240,7 +357,7 @@ class PFSMixin:
         >>>     client.delete_file(c, "/dir/delete_me.txt")
         >>>     client.put_file_bytes(c, "/new_file.txt", b"DATA")
         """
-        commit = self.start_commit(repo_name, branch, parent, description)
+        commit = self.start_commit(repo_name, branch, parent, description, project_name)
         try:
             yield commit
         finally:
@@ -310,6 +427,8 @@ class PFSMixin:
         reverse: bool = False,
         all: bool = False,
         origin_kind: pfs_proto.OriginKind = pfs_proto.OriginKind.USER,
+        started_time: datetime = None,
+        project_name: str = None,
     ) -> Union[Iterator[pfs_proto.CommitInfo], Iterator[pfs_proto.CommitSetInfo]]:
         """Lists commits.
 
@@ -341,6 +460,9 @@ class PFSMixin:
             An enum that specifies how a subcommit originated. Returns only
             subcommits of this enum type. Only impacts results if `repo_name`
             is specified.
+        started_time : datetime
+        project_name : str, optional
+            The name of the project containing the repo.
 
         Returns
         -------
@@ -360,13 +482,21 @@ class PFSMixin:
 
         .. # noqa: W505
         """
+        project = pfs_proto.Project(name=project_name) if project_name else None
         if repo_name is not None:
+            if started_time is not None:
+                started_time = timestamp_pb2.Timestamp.FromDatetime(started_time)
             req = pfs_proto.ListCommitRequest(
-                repo=pfs_proto.Repo(name=repo_name, type="user"),
+                repo=pfs_proto.Repo(
+                    name=repo_name,
+                    type="user",
+                    project=project,
+                ),
                 number=number,
                 reverse=reverse,
                 all=all,
                 origin_kind=origin_kind,
+                started_time=started_time,
             )
             if to_commit is not None:
                 req.to = commit_from(to_commit)
@@ -374,7 +504,11 @@ class PFSMixin:
                 req.from_ = commit_from(from_commit)
             return self._req(Service.PFS, "ListCommit", req=req)
         else:
-            return self._req(Service.PFS, "ListCommitSet")
+            return self._req(
+                Service.PFS,
+                "ListCommitSet",
+                project=project,
+            )
 
     def squash_commit(self, commit_id: str) -> None:
         """Squashes a commit into its parent.
@@ -441,6 +575,7 @@ class PFSMixin:
         state: pfs_proto.CommitState = pfs_proto.CommitState.STARTED,
         all: bool = False,
         origin_kind: pfs_proto.OriginKind = pfs_proto.OriginKind.USER,
+        project_name: str = None,
     ) -> Iterator[pfs_proto.CommitInfo]:
         """Returns all commits on the branch and then listens for new commits
         that are created.
@@ -463,6 +598,8 @@ class PFSMixin:
         origin_kind : {pfs_proto.OriginKind.USER, pfs_proto.OriginKind.AUTO, pfs_proto.OriginKind.FSCK, pfs_proto.OriginKind.ALIAS}, optional
             An enum that specifies how a commit originated. Returns only
             commits of this enum type. (Default value = ``pfs_proto.OriginKind.USER``)
+        project_name : str
+            The name of the project.
 
         Returns
         -------
@@ -479,7 +616,9 @@ class PFSMixin:
 
         .. # noqa: W505
         """
-        repo = pfs_proto.Repo(name=repo_name, type="user")
+        repo = pfs_proto.Repo(
+            name=repo_name, type="user", project=pfs_proto.Project(name=project_name)
+        )
         req = pfs_proto.SubscribeCommitRequest(
             repo=repo,
             branch=branch,
@@ -502,6 +641,7 @@ class PFSMixin:
         provenance: List[pfs_proto.Branch] = None,
         trigger: pfs_proto.Trigger = None,
         new_commit: bool = False,
+        project_name: str = None,
     ) -> None:
         """Creates a new branch.
 
@@ -522,6 +662,8 @@ class PFSMixin:
         new_commit : bool, optional
             If true and `head_commit` is specified, uses a different commit ID
             for head than `head_commit`.
+        project_name : str
+            The name of the project.
 
         Examples
         --------
@@ -541,7 +683,12 @@ class PFSMixin:
             Service.PFS,
             "CreateBranch",
             branch=pfs_proto.Branch(
-                repo=pfs_proto.Repo(name=repo_name, type="user"), name=branch_name
+                name=branch_name,
+                repo=pfs_proto.Repo(
+                    name=repo_name,
+                    type="user",
+                    project=pfs_proto.Project(name=project_name),
+                ),
             ),
             head=commit_from(head_commit),
             provenance=provenance,
@@ -549,7 +696,12 @@ class PFSMixin:
             new_commit_set=new_commit,
         )
 
-    def inspect_branch(self, repo_name: str, branch_name: str) -> pfs_proto.BranchInfo:
+    def inspect_branch(
+        self,
+        repo_name: str,
+        branch_name: str,
+        project_name: str = None,
+    ) -> pfs_proto.BranchInfo:
         """Inspects a branch.
 
         Parameters
@@ -558,6 +710,8 @@ class PFSMixin:
             The name of the repo.
         branch_name : str
             The name of the branch.
+        project_name : str
+            The name of the project.
 
         Returns
         -------
@@ -568,12 +722,20 @@ class PFSMixin:
             Service.PFS,
             "InspectBranch",
             branch=pfs_proto.Branch(
-                repo=pfs_proto.Repo(name=repo_name, type="user"), name=branch_name
+                name=branch_name,
+                repo=pfs_proto.Repo(
+                    name=repo_name,
+                    type="user",
+                    project=pfs_proto.Project(name=project_name),
+                ),
             ),
         )
 
     def list_branch(
-        self, repo_name: str, reverse: bool = False
+        self,
+        repo_name: str,
+        reverse: bool = False,
+        project_name: str = None,
     ) -> Iterator[pfs_proto.BranchInfo]:
         """Lists the active branch objects in a repo.
 
@@ -583,6 +745,8 @@ class PFSMixin:
             The name of the repo.
         reverse : bool, optional
             If true, returns branches oldest to newest.
+        project_name : str
+            The name of the project.
 
         Returns
         -------
@@ -596,12 +760,20 @@ class PFSMixin:
         return self._req(
             Service.PFS,
             "ListBranch",
-            repo=pfs_proto.Repo(name=repo_name, type="user"),
+            repo=pfs_proto.Repo(
+                name=repo_name,
+                type="user",
+                project=pfs_proto.Project(name=project_name),
+            ),
             reverse=reverse,
         )
 
     def delete_branch(
-        self, repo_name: str, branch_name: str, force: bool = False
+        self,
+        repo_name: str,
+        branch_name: str,
+        force: bool = False,
+        project_name: str = None,
     ) -> None:
         """Deletes a branch, but leaves the commits themselves intact. In other
         words, those commits can still be accessed via commit IDs and other
@@ -615,12 +787,19 @@ class PFSMixin:
             The name of the branch.
         force : bool, optional
             If true, forces the branch deletion.
+        project_name : str
+            The name of the project.
         """
         self._req(
             Service.PFS,
             "DeleteBranch",
             branch=pfs_proto.Branch(
-                repo=pfs_proto.Repo(name=repo_name, type="user"), name=branch_name
+                name=branch_name,
+                repo=pfs_proto.Repo(
+                    name=repo_name,
+                    type="user",
+                    project=pfs_proto.Project(name=project_name),
+                ),
             ),
             force=force,
         )
@@ -731,6 +910,7 @@ class PFSMixin:
         recursive: bool = False,
         datum: str = None,
         append: bool = False,
+        concurrency: int = 0,
     ) -> None:
         """Uploads a PFS file using the content found at a URL. The URL is sent
         to the server which performs the request.
@@ -751,6 +931,9 @@ class PFSMixin:
         append : bool, optional
             If true, appends the data to the file(s) specified at `path`, if
             they already exist. Otherwise, overwrites them.
+        concurrency : int
+            The maximum number of threads used to complete the request.
+            Defaults to 50.
 
         Examples
         --------
@@ -771,6 +954,7 @@ class PFSMixin:
                 recursive=recursive,
                 datum=datum,
                 append=append,
+                concurrency=concurrency,
             )
 
     def copy_file(
@@ -841,7 +1025,6 @@ class PFSMixin:
             Specifies an object storage URL that the file will be uploaded to.
         offset : int, optional
             Allows file read to begin at `offset` number of bytes.
-
         Returns
         -------
         PFSFile
@@ -928,7 +1111,9 @@ class PFSMixin:
         commit: Union[tuple, dict, Commit, pfs_proto.Commit],
         path: str,
         datum: str = None,
-        details: bool = False,
+        pagination_marker: pfs_proto.File = None,
+        number: int = None,
+        reverse: bool = False,
     ) -> Iterator[pfs_proto.FileInfo]:
         """Lists the files in a directory.
 
@@ -940,8 +1125,15 @@ class PFSMixin:
             The path to the directory.
         datum : str, optional
             A tag that filters the files.
-        details : bool, optional
-            Unused.
+        pagination_marker:
+            Marker for pagination. If set, the files that come after the marker
+            in lexicographical order will be returned. If reverse is also set,
+            the files that come before the marker in lexicographical order will
+            be returned.
+        number : int, optional
+            Number of files to return
+        reverse : bool, optional
+            If true, return files in reverse order
 
         Returns
         -------
@@ -956,6 +1148,9 @@ class PFSMixin:
             Service.PFS,
             "ListFile",
             file=pfs_proto.File(commit=commit_from(commit), path=path, datum=datum),
+            paginationMarker=pagination_marker,
+            number=number,
+            reverse=reverse,
         )
 
     def walk_file(
@@ -963,6 +1158,9 @@ class PFSMixin:
         commit: Union[tuple, dict, Commit, pfs_proto.Commit],
         path: str,
         datum: str = None,
+        pagination_marker: pfs_proto.File = None,
+        number: int = None,
+        reverse: bool = False,
     ) -> Iterator[pfs_proto.FileInfo]:
         """Walks over all descendant files in a directory.
 
@@ -974,6 +1172,15 @@ class PFSMixin:
             The path to the directory.
         datum : str, optional
             A tag that filters the files.
+        pagination_marker:
+            Marker for pagination. If set, the files that come after the marker
+            in lexicographical order will be returned. If reverse is also set,
+            the files that come before the marker in lexicographical order will
+            be returned.
+        number : int, optional
+            Number of files to return
+        reverse : bool, optional
+            If true, return files in reverse order
 
         Returns
         -------
@@ -988,10 +1195,16 @@ class PFSMixin:
             Service.PFS,
             "WalkFile",
             file=pfs_proto.File(commit=commit_from(commit), path=path, datum=datum),
+            paginationMarker=pagination_marker,
+            number=number,
+            reverse=reverse,
         )
 
     def glob_file(
-        self, commit: Union[tuple, dict, Commit, pfs_proto.Commit], pattern: str
+        self,
+        commit: Union[tuple, dict, Commit, pfs_proto.Commit],
+        pattern: str,
+        path_range: pfs_proto.PathRange = None,
     ) -> Iterator[pfs_proto.FileInfo]:
         """Lists files that match a glob pattern.
 
@@ -1012,7 +1225,11 @@ class PFSMixin:
         >>> files = list(client.glob_file(("foo", "master"), "/*.txt"))
         """
         return self._req(
-            Service.PFS, "GlobFile", commit=commit_from(commit), pattern=pattern
+            Service.PFS,
+            "GlobFile",
+            commit=commit_from(commit),
+            pattern=pattern,
+            path_range=path_range,
         )
 
     def delete_file(
@@ -1168,7 +1385,7 @@ class PFSMixin:
 
         return True
 
-    def mount(self, mount_dir: str, repos: List[str] = []) -> None:
+    def mount(self, mount_dir: str, repos: List[str] = None) -> None:
         """Mounts Pachyderm repos locally.
 
         Parameters
@@ -1192,6 +1409,7 @@ class PFSMixin:
         --------
         >>> client.mount("dir_a", ["repo1", "repo2@staging"])
         """
+        repos = repos or []
         check_pachctl()
         Path(mount_dir).mkdir(parents=True, exist_ok=True)
 
@@ -1370,6 +1588,7 @@ class ModifyFileClient:
         datum: str = None,
         append: bool = False,
         recursive: bool = False,
+        concurrency: int = 0,
     ) -> None:
         """Uploads a PFS File from the content found at a URL. The URL is
         sent to the server which performs the request.
@@ -1388,6 +1607,9 @@ class ModifyFileClient:
         recursive : bool, optional
             If true, allows for recursive scraping on some types URLs, for
             example on s3:// URLs
+        concurrency : int
+            The maximum number of threads used to complete the request.
+            Defaults to 50.
         """
         self._ops.append(
             _AtomicModifyFileURLOp(
@@ -1396,6 +1618,7 @@ class ModifyFileClient:
                 datum=datum,
                 append=append,
                 recursive=recursive,
+                concurrency=concurrency,
             )
         )
 
@@ -1514,11 +1737,13 @@ class _AtomicModifyFileURLOp(_AtomicOp):
         datum: str = None,
         append: bool = False,
         recursive: bool = False,
+        concurrency: int = 0,
     ):
         super().__init__(path, datum)
         self.url = url
         self.recursive = recursive
         self.append = append
+        self.concurrency = concurrency
 
     def reqs(self) -> Iterator[pfs_proto_pb.ModifyFileRequest]:
         if not self.append:
@@ -1530,6 +1755,7 @@ class _AtomicModifyFileURLOp(_AtomicOp):
                 url=pfs_proto_pb.AddFile.URLSource(
                     URL=self.url,
                     recursive=self.recursive,
+                    concurrency=self.concurrency,
                 ),
             ),
         )
